﻿# Copyright (c) 2001-2013 WorldViz LLC.
# All rights reserved.
import viz
import vizmat

#Glvoe types
GLOVE_5DT	= 1
CYBERGLOVE	= 3
POWERGLOVE  = 5

#Finger definitions
THUMB	= 0
INDEX	= 1
MIDDLE	= 2
RING	= 3
PINKY	= 4

#Joint definitions
THUMB_1		= 0
THUMB_2		= 1
THUMB_3		= 2
INDEX_1		= 3
INDEX_2		= 4
INDEX_3		= 5
MIDDLE_1	= 6
MIDDLE_2	= 7
MIDDLE_3	= 8
RING_1		= 9
RING_2		= 10
RING_3		= 11
PINKY_1		= 12
PINKY_2		= 13
PINKY_3		= 14
WRIST		= 15
NUM_JOINTS	= 16

#Gesture definitions
GESTURE_UNKNOWN					= -1
GESTURE_FIST					= 0
GESTURE_INDEX_FINGER			= 1
GESTURE_MIDDLE_FINGER			= 2
GESTURE_TWO_FINGER				= 3
GESTURE_RING_FINGER				= 4
GESTURE_RING_INDEX_FINGER		= 5
GESTURE_RING_MIDDLE_FINGER		= 6
GESTURE_THREE_FINGER			= 7
GESTURE_LITTLE_FINGER			= 8
GESTURE_INDEX_LITTLE_FINGER		= 9
GESTURE_LITTLE_MIDDLE_FINGER	= 10
GESTURE_NOT_RING_FINGER			= 11
GESTURE_LITTLE_RING_FINGER		= 12
GESTURE_NOT_MIDDLE_FINGER		= 13
GESTURE_NOT_INDEX_FINGER		= 14
GESTURE_FLAT_HAND				= 15

#Create gesture event
HAND_GESTURE_EVENT = viz.getEventID('HandGestureEvent')

#List to create gestures
_GestureRot = {}
_GestureRot[GESTURE_FIST]					= (90,90,90,90)
_GestureRot[GESTURE_INDEX_FINGER]			= (0,90,90,90)
_GestureRot[GESTURE_MIDDLE_FINGER]			= (90,0,90,90)
_GestureRot[GESTURE_TWO_FINGER]				= (0,0,90,90)
_GestureRot[GESTURE_RING_FINGER]			= (90,90,0,90)
_GestureRot[GESTURE_RING_INDEX_FINGER]		= (0,90,0,90)
_GestureRot[GESTURE_RING_MIDDLE_FINGER]		= (90,0,0,90)
_GestureRot[GESTURE_THREE_FINGER]			= (0,0,0,90)
_GestureRot[GESTURE_LITTLE_FINGER]			= (90,90,90,0)
_GestureRot[GESTURE_INDEX_LITTLE_FINGER]	= (0,90,90,0)
_GestureRot[GESTURE_LITTLE_MIDDLE_FINGER]	= (90,0,90,0)
_GestureRot[GESTURE_NOT_RING_FINGER]		= (0,0,90,0)
_GestureRot[GESTURE_LITTLE_RING_FINGER]		= (90,90,0,0)
_GestureRot[GESTURE_NOT_MIDDLE_FINGER]		= (0,90,0,0)
_GestureRot[GESTURE_NOT_INDEX_FINGER]		= (90,0,0,0)
_GestureRot[GESTURE_FLAT_HAND]				= (0,0,0,0)

class Gesture(object):
	"""Base class for implementing custom gestures"""
	def apply(self,model,weight):
		"""Apply gesture to model"""
		pass

class JointAngleGesture(Gesture):
	"""Custom gesture that will apply specified joint angles"""
	def __init__(self,jointAngles):
		"""jointAngles is a dictionary that maps a joint index to the angle value for the gesture"""
		Gesture.__init__(self)

		#Compile list of joint angles from dictionary (exclude wrist joint)
		self.jointData = []
		for x in range(NUM_JOINTS-1):
			if x <= THUMB_3:
				self.jointData.append( ( 0 , 0 , -1 , jointAngles.get(x,0.0) ) )
			else:
				self.jointData.append( ( 0 , -1 , 0 , jointAngles.get(x,0.0) ) )

	def apply(self,model,weight):
		"""Apply gesture to model"""
		getJoint = model.getJointBone
		for joint,(x,y,z,deg) in enumerate(self.jointData):
			getJoint(joint).setAxisAngle(x,y,z,deg*weight)

class FingerAngleGesture(JointAngleGesture):
	"""Custom gesture that will apply specified finger angles"""
	def __init__(self,fingerAngles):
		"""fingerAngles is a list of angles for all 5 finger joints"""

		jointAngles = {	THUMB_2 : fingerAngles[0]
						,THUMB_3 : fingerAngles[0]
						,INDEX_1 : fingerAngles[1]
						,INDEX_2 : fingerAngles[1]
						,MIDDLE_1 : fingerAngles[2]
						,MIDDLE_2 : fingerAngles[2]
						,RING_1 : fingerAngles[3]
						,RING_2 : fingerAngles[3]
						,PINKY_1 : fingerAngles[4]
						,PINKY_2 : fingerAngles[4] }
		JointAngleGesture.__init__(self,jointAngles)

class PinchGesture(JointAngleGesture):
	"""Class for pinch related gestures"""
	pass

class OpenGesture(JointAngleGesture):
	"""Class for open hand gestures"""
	pass

def add(sensor,type = GLOVE_5DT, **kw):
	"""Create a hand from the sensor"""
	return HandModel(sensor,**kw)

#Define pinch gesture
GESTURE_PINCH_INDEX = JointAngleGesture({THUMB_1:30,THUMB_2:40,THUMB_3:35,INDEX_1:35,INDEX_2:66,INDEX_3:30})




"""
Classes that implement custom input sensor to hand object
"""

class InputSensor(object):
	"""Base class for implementing a generic input sensor for Hand objects"""
	def __init__(self):
		pass

	def valid(self):
		"""Used to emulate standard Vizard object"""
		return True

	def getData(self):
		"""Return current sensor data"""
		return None

	def createHandRenderer(self,*args,**kw):
		"""Create object that will apply input sensor data to a hand model"""
		raise NotImplementedError

class MouseSensor(InputSensor):
	"""Class for using mouse button to control hand gesture"""
	def __init__(self,button,gestureDown=GESTURE_FIST,gestureUp=GESTURE_FLAT_HAND):
		InputSensor.__init__(self)
		self.button = button
		self.gestureDown = (gestureDown,True,True)
		self.gestureUp = (gestureUp,False,False)

	def getData(self):
		if viz.mouse.getState(True) & self.button:
			return self.gestureDown
		return self.gestureUp

	def createHandRenderer(self,*args,**kw):
		"""Create object that will apply input sensor data to a hand model"""
		return _InputDeviceRenderer(*args,**kw)

class KeySensor(InputSensor):
	"""Class for using key press to control hand gesture"""
	def __init__(self,key,gestureDown=GESTURE_FIST,gestureUp=GESTURE_FLAT_HAND):
		InputSensor.__init__(self)
		self.key = key
		self.gestureDown = (gestureDown,True,True)
		self.gestureUp = (gestureUp,False,False)

	def getData(self):
		if viz.key.isDown(self.key,True):
			return self.gestureDown
		return self.gestureUp

	def createHandRenderer(self,*args,**kw):
		"""Create object that will apply input sensor data to a hand model"""
		return _InputDeviceRenderer(*args,**kw)

class CompositeInputSensor(InputSensor):
	"""Class for combining multiple keyboard/mouse sensors for controlling hand gestures"""
	def __init__(self,sensors,gestureUp=GESTURE_FLAT_HAND):
		InputSensor.__init__(self)
		self.sensors = sensors
		self.gestureUp = (gestureUp,False,False)

	def getData(self):
		for sensor in self.sensors:
			data = sensor.getData()
			if data[2]:
				return data
		return self.gestureUp

	def createHandRenderer(self,*args,**kw):
		"""Create object that will apply input sensor data to a hand model"""
		return _InputDeviceRenderer(*args,**kw)

class MultiInputSensor(InputSensor):
	"""
	This class is more powerful than the standard sensor objects. It can perform open hand,
	pinch, and fist gestures. Each of the gesture triggers can be specified using a list of keyboard
	or mouse events. You can also do your own vizact.on* calls to the gesture methods below and
	activate the gestures using anything that can call functions for on and off events. So you
	could use an exotic input device that is read in via vizact.onsensor*, etc. Also, if you want
	to have a permanently enabled gesture if you have no input, then call startFist/startPunch
	to activate it, and it will stay on permanently. Note that this code controls the actual gesture
	executed, so you cannot change the shape of the hand for each state, it is hard-coded intentionally.

	@param fistButtons: An array of keyboard and mouse events that will trigger a fist gesture
	@param pinchButtons: An array of keyboard and mouse events that will trigger a pinch gesture
	@param avatar: Set the avatar flag to true if you are dealing with a VizAvatar instead of a hand, the rotations for the pinch/fist are different in this case
	"""
	def __init__(self,fistButtons=[], pinchButtons=[], avatar=False):
		InputSensor.__init__(self)
		self.fistButtons = fistButtons
		self.pinchButtons = pinchButtons

		# Define the gestures that will represent open, pinch, and fist gestures
		if avatar: # Override the angles to make the pinch work for a LiveCharacter
			pinch = PinchGesture({THUMB_1:25,THUMB_2:20,THUMB_3:15,INDEX_1:20,INDEX_2:95,INDEX_3:0,MIDDLE_1:90,MIDDLE_2:90,RING_1:90,RING_2:90,PINKY_1:90,PINKY_2:90}) # Values calculated to cause a pinch with a character, cannot use standard py values
			open  = OpenGesture({THUMB_1:25,THUMB_2:20,THUMB_3:15,INDEX_1:00,INDEX_2:70,INDEX_3:0,MIDDLE_1:90,MIDDLE_2:90,RING_1:90,RING_2:90,PINKY_1:90,PINKY_2:90}) # Almost a pinch but just open a little bit, the finger will move in when the grab occurs
		else:
			pinch = PinchGesture({THUMB_1:30,THUMB_2:40,THUMB_3:35,INDEX_1:35,INDEX_2:66,INDEX_3:30,MIDDLE_1:90,MIDDLE_2:90,RING_1:90,RING_2:90,PINKY_1:90,PINKY_2:90}) # Standard pinch values from py
			open  = OpenGesture({THUMB_1:30,THUMB_2:40,THUMB_3:35,INDEX_1:00,INDEX_2:66,INDEX_3:30,MIDDLE_1:90,MIDDLE_2:90,RING_1:90,RING_2:90,PINKY_1:90,PINKY_2:90}) # Almost a pinch but just open a little bit, the finger will move in when the grab occurs

		self.returnFist  = (GESTURE_FIST,True,True) # The standard version of GESTURE_FIST is fine to just use as is
		self.returnOpen  = (open,False,False)
		self.returnPinch  = (pinch,True,True)
		self.countFist  = 0
		self.countPinch = 0

		# If you push a keyboard key and hold it, you will get multiple key down events but only one key up event
		# So we need to poll the keyboard each time we do getData(), this is how the KeySensor code works as well
		self.fistPollKeys = []
		self.pinchPollKeys = []
		# The mouse seems to have problems when it leaves the window while a button is down, when the mouse returns
		# it will not register an up event. So we need to use a polling mechanism for this as well
		self.fistPollMouse = []
		self.pinchPollMouse = []

		# Set up fist callbacks for all the mouse and keyboard events
		for key in fistButtons:
			if key in [ viz.MOUSEBUTTON_LEFT, viz.MOUSEBUTTON_MIDDLE, viz.MOUSEBUTTON_RIGHT]:
				self.fistPollMouse.append(key)
			else:
				self.fistPollKeys.append(key)

		# Set up pinch callbacks for all the mouse and keyboard events
		for key in pinchButtons:
			if key in [ viz.MOUSEBUTTON_LEFT, viz.MOUSEBUTTON_MIDDLE, viz.MOUSEBUTTON_RIGHT]:
				self.pinchPollMouse.append(key)
			else:
				self.pinchPollKeys.append(key)

	def startFist(self, *args, **kw):
		"""Method which can be used to manually start a fist gesture, useful for manually interfacing with exotic input devices"""
		self.countFist += 1
	def endFist(self, *args, **kw):
		"""Method which can be used to manually end a fist gesture, useful for manually interfacing with exotic input devices"""
		self.countFist -= 1
	def startPinch(self, *args, **kw):
		"""Method which can be used to manually start a pinch gesture, useful for manually interfacing with exotic input devices"""
		self.countPinch += 1
	def endPinch(self, *args, **kw):
		"""Method which can be used to manually end a pinch gesture, useful for manually interfacing with exotic input devices"""
		self.countPinch -= 1

	def getData(self):
		"""Calculate the state of the hand based on all the inputs"""
		fistnum = self.countFist
		mstate = viz.mouse.getState(immediate=True)
		for each in self.fistPollMouse:
			if each & mstate:
				fistnum += 1
		for each in self.fistPollKeys:
			if viz.key.isDown(each):
				fistnum += 1
		pinchnum = self.countPinch
		for each in self.pinchPollMouse:
			if each & mstate:
				pinchnum += 1
		for each in self.pinchPollKeys:
			if viz.key.isDown(each):
				pinchnum += 1
		if fistnum > 0:
			return self.returnFist
		elif pinchnum > 0:
			return self.returnPinch
		else:
			return self.returnOpen

	def createHandRenderer(self,*args,**kw):
		"""Internal method to create object that will apply input sensor data to a hand model"""
		return _InputDeviceRenderer(*args,**kw)

class AvatarSensor(InputSensor):
	"""Class for using existing avatar to control hand gesture"""
	def __init__(self,avatar,left=False,fingerCloseAngle=90.0):
		InputSensor.__init__(self)

		joint_names = """	Bip01 %s Finger0
							Bip01 %s Finger01
							Bip01 %s Finger1
							Bip01 %s Finger11
							Bip01 %s Finger2
							Bip01 %s Finger21
							Bip01 %s Finger3
							Bip01 %s Finger31
							Bip01 %s Finger4
							Bip01 %s Finger41 """

		fingertip_names = """	Bip01 %s Finger0Nub
								Bip01 %s Finger1Nub
								Bip01 %s Finger2Nub
								Bip01 %s Finger3Nub
								Bip01 %s Finger4Nub """

		#Determine which hand we are getting gesture data from
		self._leftHand = left
		if self._leftHand:
			handCharacter = 'L'
			self._quatSign = -1.0
		else:
			handCharacter = 'R'
			self._quatSign = 1.0

		#Get list of all inner and outer joints for the 4 main fingers
		joint_names = [ n.strip() for n in joint_names.replace('%s',handCharacter).split('\n') ]
		self._joints = []
		self._joints.append( ( avatar.getBone( joint_names[2] ) , avatar.getBone( joint_names[3] ) ) )
		self._joints.append( ( avatar.getBone( joint_names[4] ) , avatar.getBone( joint_names[5] ) ) )
		self._joints.append( ( avatar.getBone( joint_names[6] ) , avatar.getBone( joint_names[7] ) ) )
		self._joints.append( ( avatar.getBone( joint_names[8] ) , avatar.getBone( joint_names[9] ) ) )

		#Save thumb joints for manual manipulation
		self._thumbJoints =  ( avatar.getBone( joint_names[0] ) , avatar.getBone( joint_names[1] ) )

		#Get list of all fingertip bones
		fingertip_names = [ n.strip() for n in fingertip_names.replace('%s',handCharacter).split('\n') ]
		self._fingertips = [ avatar.getBone(n) for n in fingertip_names ]

		#Get root hand bone
		self._handroot = avatar.getBone('Bip01 ' + handCharacter + ' Hand')

		#Compute rotation offset for root hand bone to align it with hand model
		#TODO: Ideally we would compute this from generic avatar, but we'll hard code it for now
		if self._leftHand:
			self._handoffset = viz.Matrix.translate(0,0,0.04) * viz.Matrix.euler([-90,41,0])
		else:
			self._handoffset = viz.Matrix.translate(0,0,0.04) * viz.Matrix.euler([90,41,0])

		#Save angle threshold for triggering finger close
		self._fingerCloseAngle = fingerCloseAngle

	def lockJoints(self):
		"""Lock avatar finger joints for manual manipulation"""
		for (inner,outer) in self._joints:
			inner.lock()
			outer.lock()
		self._thumbJoints[0].lock()
		self._thumbJoints[1].lock()

	def unlockJoints(self):
		"""Unlock avatar finger joints and let animation control fingers"""
		for (inner,outer) in self._joints:
			inner.unlock()
			outer.unlock()
		self._thumbJoints[0].unlock()
		self._thumbJoints[1].unlock()

	def applyModelToAvatar(self,model):
		"""Apply joint data of hand model to avatar"""

		#Bring to local scope
		modelJoint = model.getJointBone
		quatSign = self._quatSign

		#Update main finger joints
		for index,(inner,outer) in enumerate(self._joints):

			baseIndex = (index+1)*3

			x,y,z,w = modelJoint(baseIndex).getQuat()
			inner.setQuat(quatSign*x,quatSign*z,-y,w)

			x,y,z,w = modelJoint(baseIndex+1).getQuat()
			outer.setQuat(quatSign*x,quatSign*z,-y,w)

		#Update thumb joints
		innerThumb,outerThumb = self._thumbJoints

		#TODO: Need to figure out correct way to map thumb joint
		innerThumb.setQuat(innerThumb.getQuat(8))
		x,y,z,w = modelJoint(0).getQuat()
		innerThumb.setQuat(x,y,-z,w,viz.REL_LOCAL)

		x,y,z,w = modelJoint(2).getQuat()
		outerThumb.setQuat(x,y,-z,w)

	def getData(self):

		gesture = 0

		#Iterate over fingers to compute gesture, ignoring thumb
		for index,(inner,outer) in enumerate(self._joints):
			jointAngle = inner.getEuler()[2] + outer.getEuler()[2]
			if jointAngle < self._fingerCloseAngle:
				gesture |= (1 << index)

		return gesture

	def createHandRenderer(self,*args,**kw):
		"""Create object that will apply input sensor data to a hand model"""
		return _AvatarGestureRenderer(self._fingertips,self._handroot,self._handoffset,*args,**kw)





"""
Classes that apply various sensor data to hand model
"""

class _HandRenderer(object):
	def __init__(self,model):
		self.model = model

	def getGesture(self,data):
		"""Return gesture given sensor data"""
		return GESTURE_FLAT_HAND

	def getFingerTip(self,index):
		"""Return the transform matrix of the finger tip"""
		return None

	def getFingerTipLinkable(self,index):
		"""Return linkable object containing transform matrix of the finger tip"""
		return None

class _InputDeviceRenderer(_HandRenderer):

	def __init__(self,*args,**kw):
		_HandRenderer.__init__(self,*args,**kw)

	def update(self,data):
		"""Update model gesture from input device"""
		self.model.setGesture(data[0],closeThumb=data[1])

	def getGesture(self,data):
		"""Return gesture of input sensor"""
		return data[0]

class _AvatarGestureRenderer(_HandRenderer):

	def __init__(self,fingertips,handroot,handoffset,*args,**kw):
		_HandRenderer.__init__(self,*args,**kw)
		self._fingertips = fingertips
		self._handroot = handroot
		self._handoffset = handoffset

		#Completely hide model from scene graph
		self.model.setMask(0)

	def update(self,data):
		"""Update model gesture from input device"""

		#Update position/rotation of hand model from avatar
		m = self._handroot.getMatrix(viz.AVATAR_WORLD)
		m.preMult(self._handoffset)
		self.model.setMatrix(m,viz.ABS_GLOBAL)

	def getGesture(self,data):
		"""Return gesture of input sensor"""
		return data

	def getFingerTip(self,index):
		"""Return the transform matrix of the finger tip"""
		return self._fingertips[index].getMatrix(viz.AVATAR_WORLD)

	def getFingerTipLinkable(self,index):
		"""Return linkable object containing transform matrix of the finger tip"""
		return self._fingertips[index]

class _5DTRenderer(_HandRenderer):

	def __init__(self,sixteenSensor,*args,**kw):
		_HandRenderer.__init__(self,*args,**kw)

		self.bones = []
		self.bones.append( self.model.getJointBone(THUMB_2) )
		self.bones.append( self.model.getJointBone(THUMB_3) )
		self.bones.append( self.model.getJointBone(INDEX_1) )
		self.bones.append( self.model.getJointBone(INDEX_2) )
		self.bones.append( self.model.getJointBone(MIDDLE_1) )
		self.bones.append( self.model.getJointBone(MIDDLE_2) )
		self.bones.append( self.model.getJointBone(RING_1) )
		self.bones.append( self.model.getJointBone(RING_2) )
		self.bones.append( self.model.getJointBone(PINKY_1) )
		self.bones.append( self.model.getJointBone(PINKY_2) )
		self.bones.append( self.model.getJointBone(THUMB_1) )

		if sixteenSensor:
			self.joints = [0,1,3,4,6,7,9,10,12,13]
			self.supportsAbduction = True
		else:
			self.joints = [0,0,1,1,2,2,3,3,4,4]
			self.supportsAbduction = False

	def update(self,data):

		#Bring objects to local scope for speed boost
		bones = self.bones
		joints = self.joints

		#Blend thumb near data between THUMB_1 and THUMB_2
		bones[10].setAxisAngle( 0.0 , 0.0 , -1.0 , data[joints[0]] * 90.0 * 0.5 )
		bones[0].setAxisAngle( 0.0 , 0.0 , -1.0 , data[joints[0]] * 90.0 * 0.5 )
		bones[1].setAxisAngle( 0.0 , 0.0 , -1.0 , data[joints[1]] * 90.0 )
		for x in xrange(2,10):
			bones[x].setAxisAngle( 0.0 , -1.0 , 0.0 , data[joints[x]] * 90.0 )

		if self.model.abduct and self.supportsAbduction:
			bones[10].setAxisAngle( 0.0 , -1.0 , 0.0 , (1.0 - data[2]) * 40.0 - 35.0, viz.REL_LOCAL) #Thumb
			bones[2].setAxisAngle( 0.0 , 0.0 , 1.0 , (1.0 - data[5]) * 20.0 - 15.0, viz.REL_PARENT ) #Index
			bones[6].setAxisAngle( 0.0 , 0.0 , -1.0 , (1.0 - data[8]) * 20.0, viz.REL_PARENT ) #Ring
			bones[8].setAxisAngle( 0.0 , 0.0 , -1.0 , (1.0 - data[11]) * 20.0, viz.REL_PARENT ) #Pinky

	def getGesture(self,data):
		"""5DT gesture is contained in last field of raw data"""
		return int(data[-1])

class _CyberGloveRenderer(_HandRenderer):

	def __init__(self,*args,**kw):
		_HandRenderer.__init__(self,*args,**kw)

		self.bones = [ self.model.getJointBone(i) for i in xrange(16) ]

	def update(self,data):

		#Bring objects to local scope for speed boost
		bones = self.bones

		bones[1].setAxisAngle( 0.0 , 0.0 , 1.0 , data[1] )
		bones[2].setAxisAngle( 0.0 , 0.0 , 1.0 , data[2] )
		for x in xrange(3,15):
			bones[x].setAxisAngle( 0.0 , 1.0 , 0.0 , data[x] )

		if self.model.abduct:
			bones[THUMB_1].setAxisAngle( 0.0 , 0.0 , 1.0 , (data[0] + 25.0) * 0.8 )
			bones[THUMB_1].setAxisAngle( 0.0 , -1.0 , 0.0 , data[15] + 5.0, viz.REL_LOCAL)
			bones[INDEX_1].setAxisAngle( 0.0 , 0.0 , 1.0 , max(data[16]*20.0,-20.0), viz.REL_PARENT )
			bones[RING_1].setAxisAngle( 0.0 , 0.0 , -1.0 , data[17] - 5.0, viz.REL_PARENT )
			bones[PINKY_1].setAxisAngle( 0.0 , 0.0 , -1.0 , data[18], viz.REL_PARENT )
		else:
			bones[THUMB_1].setAxisAngle( 0.0 , 0.0 , 1.0 , data[0] * 0.52 )

		if self.model.wrist:
			bones[WRIST].setAxisAngle( 0.0 , 1.0 , 0.0 , data[20] - 30.0 )
			bones[WRIST].setAxisAngle( 0.0 , 0.0 , 1.0 , data[21]*0.5, viz.REL_LOCAL)

	def getGesture(self,data):
		"""CyberGlove gesture is contained in last field of raw data"""
		return int(data[-1])


class LeftBoneWrapper(viz.VizBone):
	"""Wraps bone of left handed model so rotations for right/left hand use the same reference frame"""

	def __init__(self,bone):
		viz.VizBone.__init__(self,bone.id)

	def setAxisAngle(self,x,y,z,deg,mode=viz.ABS_PARENT):
		"""Set axis angle of bone"""
		x,y,z,w = vizmat.AxisAngleToQuat(x,y,z,deg)
		self.setQuat(x,y,z,w,mode)

	def setQuat(self,x,y,z,w,mode=viz.ABS_PARENT):
		"""Set quat of bone"""
		viz.VizBone.setQuat(self,x,-y,-z,w,mode)

	def getQuat(self,mode=viz.ABS_PARENT):
		"""Get quat of bone"""
		x,y,z,w = viz.VizBone.getQuat(self,mode)
		return [x,-y,-z,w]

class HandModel(viz.VizAvatar,viz.EventClass):
	"""Graphical reprentation of a hand. Can attach a sensor to update automatically"""

	def __init__(self,sensor=None,type=0,file=None,left=False,abduct=True,wrist=False,priority=viz.PRIORITY_PLUGINS+1,**kw):

		#Determine hand model file to use
		if file is None:
			if left:
				file = 'hand_left.cfg'
			else:
				file = 'hand.cfg'

		#Initialize base class
		hand = viz.add(file,**kw)
		viz.VizAvatar.__init__(self,hand.id)
		viz.EventClass.__init__(self)

		self.__enabled = True
		self.__lastgesture = GESTURE_UNKNOWN
		self.__handSensor = None
		self._leftHanded = left
		self.abduct = abduct
		self.wrist = wrist
		self._fingerTipNodes = [None,None,None,None,None]

		#Check if sensor is valid
		if sensor is not None and sensor.valid():
			self.__sensor = sensor
		else:
			self.__sensor = None

		#Compile list of finger bones
		self._fingers = []
		self._fingers.append(self.getBone('END thumb'))
		self._fingers.append(self.getBone('END index'))
		self._fingers.append(self.getBone('END middle'))
		self._fingers.append(self.getBone('END ring'))
		self._fingers.append(self.getBone('END little'))

		#Compile list of finger joint bones
		jointNames = [	'bone thumb ROOT'
						,'bone thumb 0-0'
						,'bone thumb 0-1'
						,'bone index 1-0'
						,'bone index 1-1'
						,'bone index 1-2'
						,'bone middle 2-0'
						,'bone middle 2-1'
						,'bone middle 2-2'
						,'bone ring 3-0'
						,'bone ring 3-1'
						,'bone ring 3-2'
						,'bone little 4-0'
						,'bone little 4-1'
						,'bone little 4-2'
						,'bone WRIST'
						]

		#Need to create wrapper for joints when using left hand, so rotations are applied correctly
		if self._leftHanded:
			self._joint = [ LeftBoneWrapper(self.getBone(name)) for name in jointNames ]
		else:
			self._joint = [ self.getBone(name) for name in jointNames ]

		if self.__sensor:

			if isinstance(self.__sensor,InputSensor):

				self.__handSensor = self.__sensor.createHandRenderer(self)

			else:

				sensorName = self.__sensor.getName().lower()

				if '5dt' in sensorName:

					sixteenSensor = len(self.__sensor.getData()) > 16
					self.__handSensor = _5DTRenderer(sixteenSensor,self)

				elif 'cyberglove' in sensorName or 'cybertouch' in sensorName:

					#Have CyberGlove return joint data in degrees (only necessary with old sensor plugin)
					if isinstance(self.__sensor,viz.VizSensor):
						self.__sensor.command(1)

					self.__handSensor = _CyberGloveRenderer(self)

		#Lock finger joints
		for joint in self._joint:
			joint.lock()

		#If hand sensor is valid, then register for update events
		if self.__handSensor is not None:
			viz.EventClass.callback(self,viz.UPDATE_EVENT,self._update,priority=priority)

	def _jointsUpdated(self):
		"""INTERNAL: Called when joints have been modified"""
		pass

	def _getFingerTipLinkable(self,index):
		"""INTERNAL: Return linkable object of finger tip"""

		#Check if input sensor provies finger tip linkable
		if self.__handSensor is not None:
			finger = self.__handSensor.getFingerTipLinkable(index)
		if finger is None:
			finger = self._fingers[index]
		return finger

	def _update(self,e):

		#Get latest data from sensor
		data = self.__sensor.getData()

		#Check if gesture has changed
		gesture = self.__handSensor.getGesture(data)
		if gesture != self.__lastgesture:

			#Create event
			e = viz.Event()
			e.hand = self
			e.gesture = gesture
			e.oldGesture = self.__lastgesture

			#Update gesture
			self.__lastgesture = gesture

			#Send event
			viz.sendEvent(HAND_GESTURE_EVENT,e)

		#Update hand sensor with latest data
		if self.__enabled:
			self.__handSensor.update(data)
			self._jointsUpdated()

	def remove(self,*args,**kw):
		"""Remove hand model"""

		#Unregister event class
		self.unregister()

		#Remove base node object
		viz.VizAvatar.remove(self,*args,**kw)

	def getJointBone(self, joint):
		"""Get handle to joint bone"""
		return self._joint[joint]

	def getFingerTip(self, index):
		"""Return the transform matrix of the finger tip"""

		#Check if input sensor provies finger tip data
		if self.__handSensor is not None:
			m = self.__handSensor.getFingerTip(index)
			if m is not None:
				return m

		#Return matrix of finger tip bones
		return self._fingers[index].getMatrix(viz.AVATAR_WORLD)

	def getFingerTipLinkable(self, index):
		"""Return linkable object that will contain transform of finger tip"""

		#Return existing linkable
		linkable = self._fingerTipNodes[index]
		if linkable is not None:
			return linkable

		#Create linkable for fingertip
		link = viz.link( self._getFingerTipLinkable(index) , viz.NullLinkable , enabled=False )
		self._fingerTipNodes[index] = link

		return link

	def getGesture(self):
		"""Returns the current gesture of the underlying sensor"""
		return self.__lastgesture

	def setGesture(self, gesture, closeThumb=False, weight=1.0):
		"""Set the gesture of the hand. Only use when in manual mode.
		If closeThumb is True, the thumb will be closed in"""

		if isinstance(gesture,Gesture):
			self.__lastgesture = gesture
			gesture.apply(self,weight)
		else:
			self.__lastgesture = gesture
			data = _GestureRot[gesture]
			if closeThumb:
				thumbAngle = 90.0
			else:
				thumbAngle = 0.0

			#Apply gesture rotation data to joints
			self._joint[INDEX_1].setAxisAngle( 0 , -1 , 0 , data[0] * weight )
			self._joint[INDEX_2].setAxisAngle( 0 , -1 , 0 , data[0] * weight )
			self._joint[MIDDLE_1].setAxisAngle( 0 , -1 , 0 , data[1] * weight )
			self._joint[MIDDLE_2].setAxisAngle( 0 , -1 , 0 , data[1] * weight )
			self._joint[RING_1].setAxisAngle( 0 , -1 , 0 , data[2] * weight )
			self._joint[RING_2].setAxisAngle( 0 , -1 , 0 , data[2] * weight )
			self._joint[PINKY_1].setAxisAngle( 0 , -1 , 0 , data[3] * weight )
			self._joint[PINKY_2].setAxisAngle( 0 , -1 , 0 , data[3] * weight )
			self._joint[THUMB_2].setAxisAngle( 0 , 0 , -1 , thumbAngle * weight )
			self._joint[THUMB_3].setAxisAngle( 0 , 0 , -1 , thumbAngle * weight )

			#Zero out remaining joints
			self._joint[INDEX_3].setQuat(0.0,0.0,0.0,1.0)
			self._joint[MIDDLE_3].setQuat(0.0,0.0,0.0,1.0)
			self._joint[RING_3].setQuat(0.0,0.0,0.0,1.0)
			self._joint[PINKY_3].setQuat(0.0,0.0,0.0,1.0)
			self._joint[THUMB_1].setQuat(0.0,0.0,0.0,1.0)

		self._jointsUpdated()
	
	def setFingers(self, th=90.0, ind=90.0, mid=90.0, ring=90.0, little=90.0):
		"""Set finger position"""

		#Apply gesture rotation data to joints
		self._joint[INDEX_1].setAxisAngle( 0 , -1 , 0 , ind )
		self._joint[INDEX_2].setAxisAngle( 0 , -1 , 0 , ind )
		self._joint[MIDDLE_1].setAxisAngle( 0 , -1 , 0 , mid )
		self._joint[MIDDLE_2].setAxisAngle( 0 , -1 , 0 , mid )
		self._joint[RING_1].setAxisAngle( 0 , -1 , 0 , ring )
		self._joint[RING_2].setAxisAngle( 0 , -1 , 0 , ring )
		self._joint[PINKY_1].setAxisAngle( 0 , -1 , 0 , little )
		self._joint[PINKY_2].setAxisAngle( 0 , -1 , 0 , little )
		self._joint[THUMB_2].setAxisAngle( 0 , 0 , -1 , th )
		self._joint[THUMB_3].setAxisAngle( 0 , 0 , -1 , th )

		#Zero out remaining joints
		self._joint[INDEX_3].setQuat(0.0,0.0,0.0,1.0)
		self._joint[MIDDLE_3].setQuat(0.0,0.0,0.0,1.0)
		self._joint[RING_3].setQuat(0.0,0.0,0.0,1.0)
		self._joint[PINKY_3].setQuat(0.0,0.0,0.0,1.0)
		self._joint[THUMB_1].setQuat(0.0,0.0,0.0,1.0)

		self._jointsUpdated()

	def getSensor(self):
		"""Returns the sensor belonging to the hand"""
		return self.__sensor

	def setSensorUpdate(self,flag):
		"""Set whether the sensor will update the hand automatically"""
		self.__enabled = flag

	def getSensorUpdate(self):
		"""Returns whether the sensor will update the hand automatically"""
		return self.__enabled

	def manual(self):
		"""Switch to manual mode. The hand will not be updated by the sensor"""
		self.__enabled = False

	def automatic(self):
		"""Switch to automatic mode. The hand will be updated by the sensor"""
		self.__enabled = True

	def leftHand(self):
		"""Make the hand left-handed"""

		#Ignore command if already using left hand model
		if self._leftHanded:
			return

		self.scale(-1,1,1)
		self.enable(viz.FLIP_POLYGON_ORDER)

	def rightHand(self):
		"""Make the hand right-handed"""

		#Ignore command if already using left hand model
		if self._leftHanded:
			return

		self.scale(1,1,1)
		self.disable(viz.FLIP_POLYGON_ORDER)

class AvatarHandModel(HandModel):
	"""Version of HandModel class that applies data to a Complete Characters avatar"""
	def __init__(self,model,*args,**kw):

		#Initialize base class
		HandModel.__init__(self,*args,**kw)

		#Create avatar sensor/renderer for reusing code for applying data
		self._avatarSensor = AvatarSensor(model,left=self._leftHanded)
		self._avatarRenderer = self._avatarSensor.createHandRenderer(self)
		self._avatarSensor.lockJoints()

	def _jointsUpdated(self):
		"""INTERNAL: Called when joints have been modified"""

		#Update renderer, which will update hand position/rotation
		self._avatarRenderer.update(None) #Renderer does not use passed in data, so just give it None

		#Update joint data of avatar
		self._avatarSensor.applyModelToAvatar(self)

	def _getFingerTipLinkable(self,index):
		"""INTERNAL: Return linkable object of finger tip"""
		return self._avatarRenderer.getFingerTipLinkable(index)

	def getFingerTip(self, index):
		"""Return the transform matrix of the finger tip"""
		return self._avatarRenderer.getFingerTip(index)
